//#define CPUTRACKER 1048576
//#define SPCTRACKER 1048576
//#define DMATRACKER
//#define HDMATRACKER
/*  ROM IMAGE LOADING MECHANISM
    Original code by Savoury SnaX & Santeri Saarimaa
    New development by Charles Bilyue'

    Supports (with or without header):
     Single-part non-interleaved ROMs (LoROM and HiROM)
     Single-part interleaved HiROMs
     Multi-part non-interleaved ROMs (LoROM and HiROM, *.1-style)
     Multi-part interleaved HiROMs (*.1-style)

    If a header exists, it is ignored

       Single-part ROM extensions
        SFC = Super FamiCom
        FIG = Pro Fighter
        SMC = Super MagiCom
        SWC = Super WildCard

        BIN = Binary

       Multi-part ROM extensions
        SF*a.058,SF*a.078 = Game Doctor
        *.1 = Other miscellaneous multi-part ROM

    TO DO:
        Interleaved LoROM loading (??? need info)
        Other memory maps
*/

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <dir.h>

#include "wrapaleg.h"   // Wrapper header for Allegro
#include "romload.h"
#include "helper.h"
#include "CPU.h"
#include "SPC.h"
#include "mem.h"
#include "sound.h"
#include "tiles.h"

char *rom_romfile=0;
char *rom_romhilo=0;
char *rom_romname=0;
char *rom_romtype=0;
char *rom_romsize=0;
char *rom_sram=0;
char *rom_country=0;

extern void Timer_Handler_Throttle(...);
unsigned char *RomAddress;      // Address of SNES ROM

// Used to determine size of file for saving/loading, and to restrict writes
//  to non-existant SRAM
unsigned long SaveRamLength=0;

int Allocate_ROM();
int NumBanks;
int ROM_format;
bool ROM_Loaded = FALSE;

unsigned long Frametime;

SNESRomInfoStruct RomInfoLo, RomInfoHi;

int LoadSRAM(char *SRAM_filename){
 if (!ROM_Loaded || !SaveRamLength) return 0;   // No SRAM

 // Return if we can't get SRAM filename
 if (CreateSRAMFilename(ROM_filename)) return 0;

 FILE *Infile = fopen(SRAM_filename, "rb");
 if (!Infile) return 0;         // Can't open file

 fread(SRAM, 1, SaveRamLength, Infile); // Read Save Ram

 return -1;
}

int SaveSRAM(char *SRAM_filename){
 if (!ROM_Loaded || !SaveRamLength) return 0;   // No SRAM

 // Return if we can't get SRAM filename
 if (CreateSRAMFilename(ROM_filename)) return 0;

 FILE *Outfile = fopen(SRAM_filename, "wb");
 if (!Outfile) return 0;        // Can't open file

 fwrite(SRAM, 1, SaveRamLength, Outfile);   // Write Save Ram

 return -1;
}

char *ROM_filename = 0;
char fn_drive[MAXDRIVE], fn_dir[MAXDIR], fn_file[MAXFILE], fn_ext[MAXEXT];
char SRAM_filename[MAXPATH];
char save_dir[MAXPATH];
char save_extension[4];

ROMFileType GetROMFileType(char *ROM_filename){
 fnsplit(ROM_filename, fn_drive, fn_dir, fn_file, fn_ext);
 if (!fn_ext) return ROMFileType_normal;
 if (!strcmp(fn_ext, ".1")) return ROMFileType_split;
/*
 if (fn_file){
  if (!strnicmp(fn_file, "SF", 2) &&
      (!strcmp(fn_ext, ".058") || !strcmp(fn_ext, ".078"))){
   printf("Game Doctor ROM detected.\n");
   return ROMFileType_gamedoctor;
  }
 }
 if (!strcmp(fn_ext, ".zip")){
  printf("Compressed ROM detected.\n");
  return ROMFileType_compressed;
 }
 if (!strcmp(fn_ext, ".rar")){
  printf("Compressed ROM detected.\n");
  return ROMFileType_compressed;
 }*/
 return ROMFileType_normal;
}

bool CreateSRAMFilename(char *ROM_filename){
 int length, slength;

 SRAM_filename[0] = 0;

 length = 1;    // 1 for the null
 fnsplit(ROM_filename, fn_drive, fn_dir, fn_file, fn_ext);
 if (!fn_file) return TRUE;
 length += strlen(fn_file);

 if (strlen(save_extension)){
  length += strlen(save_extension);
  if (save_extension[0] != '.') ++length; // One for leading period
 }

 slength = strlen(save_dir);
 if (slength){
  length += slength;
  if (save_dir[slength - 1] != '/' && save_dir[slength - 1] != '\\')
   ++length;    // One for trailing slash
 } else {
  if (strlen(fn_drive)){
   length += strlen(fn_drive);
  }

  slength = strlen(fn_dir);
  if (slength){
   length += slength;
   if (fn_dir[slength - 1] != '/' && fn_dir[slength - 1] != '\\')
    ++length;    // One for trailing slash
  }
 }

 if (length > MAXPATH) return TRUE;

 if (strlen(save_dir)){
  strcat(SRAM_filename, save_dir);
  if (save_dir[slength - 1] != '/' && save_dir[slength - 1] != '\\')
   strcat(SRAM_filename, "/");  // Add missing trailing slash
 } else {
  if (strlen(fn_drive)){
   strcat(SRAM_filename, fn_drive);
  }

  if (strlen(fn_dir)){
   strcat(SRAM_filename, fn_dir);
   if (fn_dir[slength - 1] != '/' && fn_dir[slength - 1] != '\\')
    strcat(SRAM_filename, "/");  // Add missing trailing slash
  }
 }

 strcat(SRAM_filename, fn_file);

 if (strlen(save_extension)){
  if (save_extension[0] != '.')
   strcat(SRAM_filename, ".");  // Add missing leading period
  strcat(SRAM_filename, save_extension);
 }

 return FALSE;
}

void DisplayRomStats(SNESRomInfoStruct *RomInfo);

#ifdef DEBUG
#ifdef HDMATRACKER
unsigned char *HDMAAddress=0;
unsigned long LastHDMA;
#endif
#ifdef DMATRACKER
unsigned char *DMAAddress=0;
unsigned long LastDMA;
#endif
#ifdef CPUTRACKER
unsigned char *InsAddress=0;
unsigned long LastIns;
#endif
#ifdef SPCTRACKER
unsigned char *SPC_InsAddress=0;
unsigned long SPC_LastIns;
#endif
#endif

// Allocates some memory used by emulation: tile caches,virtual output screen
int Setup_Memory(){
// - Align       = 1
// - 2bpl tiles  = 4
// - 4bpl tiles  = 2
// - 8bpl tiles  = 1
//         Total = 8

 static unsigned char *AllocRAMAddress=0;
 unsigned char *temp;

#ifdef DEBUG
#ifdef HDMATRACKER
 HDMAAddress=(unsigned char *)malloc(1024);
#endif
#ifdef DMATRACKER
 DMAAddress=(unsigned char *)malloc(1024);
#endif
#ifdef CPUTRACKER
#if CPUTRACKER >= 16
 InsAddress=(unsigned char *)malloc(CPUTRACKER);
#else
 InsAddress=(unsigned char *)malloc(5120);
#endif
#endif
#ifdef SPCTRACKER
 SPC_InsAddress=(unsigned char *)malloc(SPCTRACKER);
#endif
#endif

 if(AllocRAMAddress) free(AllocRAMAddress);

 AllocRAMAddress=(unsigned char *) malloc(8 * 65536);
 if (!AllocRAMAddress) return 8 * 65536;

 temp = // Force 64k alignment
  (unsigned char *)(((unsigned long)(AllocRAMAddress+0xFFFF))&~0xFFFF);

 TileCache2 = temp;
 temp += 0x40000;
 TileCache4 = temp;
 temp += 0x20000;
 TileCache8 = temp;
 temp += 0x10000;

 return 0;
}

// NumBanks is 64k blocks
int Allocate_ROM(){
 // ROM is now dynamically allocated. This is to add large ROM support
 // without raising RAM requirements when using smaller ROMs.
 static unsigned char *AllocROMAddress=0;

 // De-allocate any previous memory
 if(AllocROMAddress) free(AllocROMAddress);

 AllocROMAddress=(unsigned char *)malloc((NumBanks+1)*65536);
 if(!AllocROMAddress) return (NumBanks+1)*65536;

 RomAddress=  // Force 64k alignment
  (unsigned char *)(((unsigned long)(AllocROMAddress+0xFFFF))&~0xFFFF);
 memset(RomAddress,0xFF,NumBanks*65536);
 return 0;
}

// Resets memory heaps
void Reset_Memory(){
#ifdef DEBUG
#ifdef HDMATRACKER
 // Reset HDMA tracking buffer
 memset(HDMAAddress,0xFF,1024); LastHDMA=0;
#endif
#ifdef DMATRACKER
 // Reset DMA tracking buffer
 memset(DMAAddress,0xFF,1024); LastDMA=0;
#endif
#ifdef CPUTRACKER
 // Reset CPU instruction tracking buffer
#if CPUTRACKER >= 16
 memset(InsAddress,0xFF,CPUTRACKER); LastIns=0;
#else
 memset(InsAddress,0,5120); LastIns=0;
#endif
#endif
#ifdef SPCTRACKER
 // Reset SPC instruction tracking buffer
 memset(SPC_InsAddress,0xFF,SPCTRACKER); SPC_LastIns=0;
#endif
#endif

 // Reset SPC address space to 0xFF
 memset(SPCRAM,0xFF,65536);

 // Reset WRAM to 0x55
 memset(WRAM,0x55,131072);

 // Reset VRAM to 0
 memset(VRAM,0,65536);
 memset(VRAM_DI,0,65536);

 // Reset tile caches to 0
 memset(TileCache2,0,262144);
 memset(TileCache4,0,131072);
 memset(TileCache8,0,65536);

 // Reset blank ROM space to 0xFF
 memset(Blank,0xFF,65536);

 // Background allocated such that there is space for 8 tiles on either side of
 // the screen, this allows the rendering to function without breaking up any
 // of the plotters!
 // Clear virtual SNES output screen
 memset(SNES_Screen8,0,(256+64)*240);

 // Clear OAM
 memset(OAM,0,512+32);
}

// Resets Save RAM
void Reset_SRAM(){
 // Reset SRAM to 0xAA
 memset(SRAM,0xAA,131072);
}

unsigned long SRAM_Mask;

inline void set_block_read_handler(int bank, int block, void (*read)(void))
{
 Read_Bank8Mapping [bank * 8 + block] = (void (*)(void)) read;
 Read_Bank8Offset [bank * 8 + block] = (void *) 0;
}

inline void set_block_read_pointer(int bank, int block, void *read)
{
 Read_Bank8Mapping [bank * 8 + block] = (void (*)(void)) 0;
 Read_Bank8Offset [bank * 8 + block] = (void *) read;
}

inline void set_block_write_handler(int bank, int block, void (*write)(void))
{
 Write_Bank8Mapping[bank * 8 + block] = (void (*)(void)) write;
 Write_Bank8Offset[bank * 8 + block] = (void *) 0;
}

inline void set_block_write_pointer(int bank, int block, void *write)
{
 Write_Bank8Mapping[bank * 8 + block] = (void (*)(void)) 0;
 Write_Bank8Offset[bank * 8 + block] = (void *) write;
}

inline void set_block_handlers(int bank, int block, void (*read)(void), void (*write)(void))
{
 set_block_read_handler(bank, block, read);
 set_block_write_handler(bank, block, write);
}

inline void set_block_pointers(int bank, int block, void *read, void *write)
{
 set_block_read_pointer(bank, block, read);
 set_block_write_pointer(bank, block, write);
}

inline void map_wram(int bank, int block)
{
 set_block_pointers(bank, block, (void *) (WRAM - (bank << 16)), (void *) (WRAM - (bank << 16)));
}

inline void map_wram_128k(int bank)
{
 for (int i = 0; i < 8; i++)
 {
  set_block_pointers(bank, i, (void *) (WRAM - (bank << 16)), (void *) (WRAM - (bank << 16)));
  set_block_pointers(bank + 1, i, (void *) (WRAM - (bank << 16)), (void *) (WRAM - (bank << 16)));
 }
}

inline void map_blank(int bank, int block)
{
 set_block_pointers(bank, block, (void *) (Blank - (bank << 16)), (void *) (Dummy - (bank << 16)));
}

inline void map_blank_32k(int bank)
{
 for (int i = 4; i < 8; i++) map_blank(bank, i);
}

inline void map_blank_64k(int bank)
{
 for (int i = 0; i < 8; i++) map_blank(bank, i);
}

inline void map_unmapped(int bank, int block)
{
 set_block_handlers(bank, block, &UNSUPPORTED_READ, &UNSUPPORTED_WRITE);
}

inline void map_unmapped_32k(int bank)
{
 for (int i = 4; i < 8; i++) map_unmapped(bank, i);
}

inline void map_unmapped_64k(int bank)
{
 for (int i = 0; i < 8; i++) map_unmapped(bank, i);
}

inline void map_sram(int bank, int block, int sram_block)
{
 set_block_read_pointer(bank, block,
  SRAM + (sram_block << 13) - (bank << 16) - (block << 13));
 set_block_write_handler(bank, block, &SRAM_WRITE);
}

inline void map_sram_2k(int bank, int block)
{
 set_block_read_pointer(bank, block, SRAM - (bank << 16) - (block << 13));
 set_block_write_handler(bank, block, &SRAM_WRITE_2k);
}

inline void map_sram_4k(int bank, int block)
{
 set_block_read_pointer(bank, block, SRAM - (bank << 16) - (block << 13));
 set_block_write_handler(bank, block, &SRAM_WRITE_4k);
}

inline void map_sram_64k(int bank, int sram_block, int mask)
{
 for (int i = 0; i < 8; i++) map_sram(bank, i, sram_block + (i & mask));
}

inline void map_sram_2k_64k(int bank)
{
 for (int i = 0; i < 8; i++) map_sram_2k(bank, i);
}

inline void map_sram_4k_64k(int bank)
{
 for (int i = 0; i < 8; i++) map_sram_4k(bank, i);
}

inline void map_rom_32k(int bank)
{
 for (int i = 4; i < 8; i++)
 {
  set_block_pointers(bank, i, (void *) (RomAddress - (bank >= 0x80 ? 0x800000 : 0)), (void *) (Dummy - (bank << 16)));
 }
}

inline void map_rom_32k_2(int bank)
{
 for (int i = 4; i < 8; i++)
 {
  set_block_pointers(bank, i, (void *) (RomAddress - (NumBanks << 16) - 0x8000 - (bank >= 0x80 ? 0x800000 : 0)), (void *) (Dummy - (bank << 16)));
 }
}

inline void map_rom_64k(int bank)
{
 for (int i = 0; i < 8; i++)
 {
  set_block_pointers(bank, i, (RomAddress - 0x400000 - (bank >= 0x80 ? 0x800000 : 0)), (void *) (Dummy - (bank << 16)));
 }
}

inline void map_ports(int bank)
{
 map_wram(bank, 0);
 set_block_handlers(bank, 1, &PPU_READ, &PPU_WRITE);
 set_block_handlers(bank, 2, &PPU_READ, &PPU_WRITE);
 map_unmapped(bank, 3);
}

void Set_LoROM_Map(void)
{

 int b;

 SRAM_Mask = SaveRamLength - 1;

 RomReadAdjust   =RomAddress;
 RomReadAdjust2  =RomAddress - (NumBanks << 16) - 0x008000;
 RomHiReadAdjust =RomAddress - 0x800000;
 RomHiReadAdjust2=RomAddress - (NumBanks << 16) - 0x808000;

 for (b = 0; b <= 0xFF; b++)
 {
  Read_BankMapping[b] = &UNSUPPORTED_READ;
  Write_BankMapping[b] = &UNSUPPORTED_WRITE;
  Read_BankOffset[2*b  ] = (void *) Blank;
  Read_BankOffset[2*b+1] = (void *) Blank;

  map_unmapped_64k(b);
 }

 for (b = 0; b <= 0x3F; b++)
 {
  Read_BankOffset[2*b       ] = (void *) WRAM;
  Read_BankOffset[2*(b+0x80)] = (void *) WRAM;

  map_ports(b);
  map_ports(b + 0x80);

  Write_BankMapping[b       ] = &WRITE_00_3F;
  Write_BankMapping[b + 0x80] = &WRITE_80_BF;

  if (b < NumBanks)
  {
   map_rom_32k(b);
   map_rom_32k(b + 0x80);
   Read_BankMapping[b       ] = &READ_00_3F_No_SRAM;
   Read_BankMapping[b + 0x80] = &READ_80_BF_No_SRAM;
   Read_BankOffset[1+2*b       ] = (void *) (RomAddress+(b<<16));
   Read_BankOffset[1+2*(b+0x80)] = (void *) (RomAddress+(b<<16));
  }
  else if (b < NumBanks * 2)
  {
   map_rom_32k_2(b);
   map_rom_32k_2(b + 0x80);
   Read_BankMapping[b       ] = &READ_00_3F_2;
   Read_BankMapping[b + 0x80] = &READ_80_BF_2;
   Read_BankOffset[1+2*b       ] = (void *) (RomAddress+((b-NumBanks)<<16)-0x8000);
   Read_BankOffset[1+2*(b+0x80)] = (void *) (RomAddress+((b-NumBanks)<<16)-0x8000);
  } else {
   map_unmapped_32k(b);
   map_unmapped_32k(b + 0x80);
   Read_BankMapping[b       ] = &BlankL_00_3F;
   Read_BankMapping[b + 0x80] = &BlankL_80_BF;
   Read_BankOffset[1+2*b       ] = (void *) Blank;
   Read_BankOffset[1+2*(b+0x80)] = (void *) Blank;
  }
 }

 for (b = 0x40; b <= 0x7F; b++)
 {
  if (b < NumBanks * 2)
  {
   map_rom_32k_2(b);
   map_rom_32k_2(b + 0x80);
   Read_BankMapping[b       ] = &READ_40_6F_2;
   Read_BankMapping[b + 0x80] = &READ_C0_FF_2;
   Read_BankOffset[1+2*b       ] = (void *) (RomAddress+((b-NumBanks)<<16)-0x8000);
   Read_BankOffset[1+2*(b+0x80)] = (void *) (RomAddress+((b-NumBanks)<<16)-0x8000);
  } else {
   map_unmapped_32k(b);
   map_unmapped_32k(b + 0x80);
   Read_BankMapping[b       ] = &BlankL_40_6F;
   Read_BankMapping[b + 0x80] = &BlankL_C0_FF;
   Read_BankOffset[1+2*b       ] = (void *) Blank;
   Read_BankOffset[1+2*(b+0x80)] = (void *) Blank;
  }
 }

 for (b = 0x70; b <= 0x7D; b++)
 {
  switch (SaveRamLength)
  {
   case 0: // No SRAM
    Write_BankMapping[b] = &UNSUPPORTED_WRITE;
    break;
   case 0x800:  // 2k SRAM
    Write_BankMapping[b] = &SRAM_WRITE_2k;
    map_sram_2k_64k(b);
    break;
   case 0x1000: // 4k SRAM
    Write_BankMapping[b] = &SRAM_WRITE_4k;
    map_sram_4k_64k(b);
    break;
   default:
    Write_BankMapping[b] = &SRAM_WRITE;
    switch (SaveRamLength)
    {
     case 0x10000: // 64k SRAM
      map_sram_64k(b, 0, 7);
      break;
     case 0x8000: // 32k SRAM
      map_sram_64k(b, 0, 3);
      break;
     case 0x4000: // 16k SRAM
      map_sram_64k(b, 0, 1);
      break;
     case 0x2000: // 8k SRAM
     default:
      map_sram_64k(b, 0, 0);
      break;
    }
  }

  Read_BankMapping[b] = SaveRamLength ? &SRAM_READ : &UNSUPPORTED_READ;
  Read_BankOffset[2*b  ] = (void *) Blank;
  Read_BankOffset[2*b+1] = (void *) Blank;
 }

 map_wram_128k(0x7E);
 for (b = 0x7E; b <= 0x7F; b++)
 {
  Read_BankMapping[b] = &RAM_READ;
  Write_BankMapping[b] = &RAM_WRITE;
  Read_BankOffset[2*b  ] = (void *) (WRAM + (b - 0x7E) * 0x10000);
  Read_BankOffset[2*b+1] = (void *) (WRAM + (b - 0x7E) * 0x10000);
 }

FILE *MAP_DUMP = fopen("map.dmp", "wb");
if (MAP_DUMP)
{
 fwrite(Read_Bank8Mapping,1,256*4*8,MAP_DUMP);
 fwrite(Read_Bank8Offset,1,256*4*8,MAP_DUMP);
 fwrite(Write_Bank8Mapping,1,256*4*8,MAP_DUMP);
 fwrite(Write_Bank8Offset,1,256*4*8,MAP_DUMP);
 fwrite(Read_BankMapping,1,256*4,MAP_DUMP);
 fwrite(Write_BankMapping,1,256*4,MAP_DUMP);
 fclose(MAP_DUMP);
}
}

void Set_HiROM_Map(void)
{

 int b;

 SRAM_Mask = SaveRamLength >= 0x2000 ? 0x1FFF : SaveRamLength - 1;

 RomReadAdjust       = RomAddress;
 HRomLinReadAdjust   = RomAddress - 0x400000;
 RomHiReadAdjust     = RomAddress - 0x800000;
 HRomHiLinReadAdjust = RomAddress - 0xC00000;

 for (b = 0; b <= 0xFF; b++)
 {
  Read_BankMapping[b] = &UNSUPPORTED_READ;
  Write_BankMapping[b] = &UNSUPPORTED_WRITE;
  Read_BankOffset[2*b  ] = (void *) Blank;
  Read_BankOffset[2*b+1] = (void *) Blank;

  map_unmapped_64k(b);
 }

 for (b = 0; b <= 0x3F; b++)
 {
  Read_BankOffset[2*b       ] = (void *) WRAM;
  Read_BankOffset[2*(b+0x80)] = (void *) WRAM;

  map_ports(b);
  map_ports(b + 0x80);

  Write_BankMapping[b       ] = &HWRITE_00_3F_No_SRAM;
  Write_BankMapping[b + 0x80] = &HWRITE_80_BF_No_SRAM;
  Write_BankMapping[b + 0x40] = &UNSUPPORTED_WRITE;
  Write_BankMapping[b + 0xC0] = &UNSUPPORTED_WRITE;

  if (b < NumBanks)
  {
   // Set 32k banks
   map_rom_32k(b);
   map_rom_32k(b + 0x80);
   Read_BankMapping[b       ] = &READ_00_3F_No_SRAM;
   Read_BankMapping[b + 0x80] = &READ_80_BF_No_SRAM;
   Read_BankOffset[1+2*b       ] = (void *) (RomAddress+(b<<16));
   Read_BankOffset[1+2*(b+0x80)] = (void *) (RomAddress+(b<<16));

   // Set 64k banks
   map_rom_64k(b + 0x40);
   map_rom_64k(b + 0xC0);
   Read_BankMapping[b + 0x40] = &HREAD_40_7D;
   Read_BankMapping[b + 0xC0] = &HREAD_C0_FF;

   Read_BankOffset[2*(b+0x40)  ] = (void *) (RomAddress+(b<<16));
   Read_BankOffset[2*(b+0x40)+1] = (void *) (RomAddress+(b<<16));
   Read_BankOffset[2*(b+0xC0)  ] = (void *) (RomAddress+(b<<16));
   Read_BankOffset[2*(b+0xC0)+1] = (void *) (RomAddress+(b<<16));

  } else {
   // Set 32k banks
   map_unmapped_32k(b);
   map_unmapped_32k(b + 0x80);
   Read_BankMapping[b       ] = &BlankH_00_3F_No_SRAM;
   Read_BankMapping[b + 0x80] = &BlankH_80_BF_No_SRAM;
   Read_BankOffset[1+2*b       ] = (void *) Blank;
   Read_BankOffset[1+2*(b+0x80)] = (void *) Blank;

   // Set 64k banks
   map_unmapped_64k(b + 0x40);
   map_unmapped_64k(b + 0xC0);
   Read_BankMapping[b + 0x40] = &BlankH_40_7D;
   Read_BankMapping[b + 0xC0] = &BlankH_C0_FF;

   Read_BankOffset[2*(b+0x40)  ] = (void *) Blank;
   Read_BankOffset[2*(b+0x40)+1] = (void *) Blank;
   Read_BankOffset[2*(b+0xC0)  ] = (void *) Blank;
   Read_BankOffset[2*(b+0xC0)+1] = (void *) Blank;

  }
 }

 switch (SaveRamLength) // Setup mapper for rest of 32k ROM banks + SRAM
 {
  case 0:       // No SRAM
   for (b = 0x30; b <= 0x3F; b++)
   {
    map_unmapped(b, 3);
    map_unmapped(b + 0x80, 3);
    if (b < NumBanks)
    {
     Read_BankMapping [b     ] = HiROM_Read_No_SRAM_ROM_30;
     Read_BankMapping [b+0x80] = HiROM_Read_No_SRAM_ROM_B0;
     Write_BankMapping[b     ] = HiROM_Write_No_SRAM_30;
     Write_BankMapping[b+0x80] = HiROM_Write_No_SRAM_B0;
    } else {
     Read_BankMapping [b     ] = HiROM_Read_No_SRAM_Blank_30;
     Read_BankMapping [b+0x80] = HiROM_Read_No_SRAM_Blank_B0;
     Write_BankMapping[b     ] = HiROM_Write_No_SRAM_30;
     Write_BankMapping[b+0x80] = HiROM_Write_No_SRAM_B0;
    }
   }

   break;
  case 0x800:   // 2k SRAM
  case 0x1000:  // 4k SRAM
  case 0x2000:  // 8k SRAM
   for (b = 0x30; b <= 0x3F; b++)
   {
    if (b < NumBanks)
    {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_ROM_30[0];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_ROM_B0[0];
    } else {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_Blank_30[0];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_Blank_B0[0];
    }

    switch (SaveRamLength)
    {
     case 0x800:   // 2k SRAM
      map_sram_2k(b, 3);
      map_sram_2k(b + 0x80, 3);
      Write_BankMapping[b     ] = &HWRITE_3X_BX_2k;
      Write_BankMapping[b+0x80] = &HWRITE_3X_BX_2k;
      break;
     case 0x1000:  // 4k SRAM
      map_sram_4k(b, 3);
      map_sram_4k(b + 0x80, 3);
      Write_BankMapping[b     ] = &HWRITE_3X_BX_4k;
      Write_BankMapping[b+0x80] = &HWRITE_3X_BX_4k;
      break;
     case 0x2000:  // 8k SRAM
      map_sram(b, 3, 0);
      map_sram(b + 0x80, 3, 0);
      Write_BankMapping[b     ] = HiROM_Write_SRAM_30[0];
      Write_BankMapping[b+0x80] = HiROM_Write_SRAM_B0[0];
    } 
   }

   break;
  case 0x4000:  // 16k SRAM
   for (b = 0x30; b <= 0x3F; b++)
   {
    map_sram(b, 3, b & 1);
    map_sram(b + 0x80, 3, b & 1);
    if (b < NumBanks)
    {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_ROM_30[b&1];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_ROM_B0[b&1];
    } else {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_Blank_30[b&1];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_Blank_B0[b&1];
    }

    Write_BankMapping[b     ] = HiROM_Write_SRAM_30[b&1];
    Write_BankMapping[b+0x80] = HiROM_Write_SRAM_B0[b&1];
   }

   break;
  case 0x8000:  // 32k SRAM
   for (b = 0x30; b <= 0x3F; b++)
   {
    map_sram(b, 3, b & 3);
    map_sram(b + 0x80, 3, b & 3);
    if (b < NumBanks)
    {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_ROM_30[b&3];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_ROM_B0[b&3];
    } else {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_Blank_30[b&3];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_Blank_B0[b&3];
    }

    Write_BankMapping[b     ] = HiROM_Write_SRAM_30[b&3];
    Write_BankMapping[b+0x80] = HiROM_Write_SRAM_B0[b&3];
   }

   break;
  case 0x10000: // 64k SRAM
   for (b = 0x30; b <= 0x3F; b++)
   {
    map_sram(b, 3, b & 7);
    map_sram(b + 0x80, 3, b & 7);
    if (b < NumBanks)
    {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_ROM_30[b&7];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_ROM_B0[b&7];
    } else {
     Read_BankMapping [b     ] = HiROM_Read_SRAM_Blank_30[b&7];
     Read_BankMapping [b+0x80] = HiROM_Read_SRAM_Blank_B0[b&7];
    }

    Write_BankMapping[b     ] = HiROM_Write_SRAM_30[b&7];
    Write_BankMapping[b+0x80] = HiROM_Write_SRAM_B0[b&7];
   }

   break;
 }

 map_wram_128k(0x7E);
 for (b = 0x7E; b <= 0x7F; b++)
 {
  Read_BankMapping[b] = &RAM_READ;
  Write_BankMapping[b] = &RAM_WRITE;
  Read_BankOffset[2*b  ] = (void *) (WRAM + (b - 0x7E) * 0x10000);
  Read_BankOffset[2*b+1] = (void *) (WRAM + (b - 0x7E) * 0x10000);
 }

FILE *MAP_DUMP = fopen("map.dmp", "wb");
if (MAP_DUMP)
{
 fwrite(Read_Bank8Mapping,1,256*4*8,MAP_DUMP);
 fwrite(Read_Bank8Offset,1,256*4*8,MAP_DUMP);
 fwrite(Write_Bank8Mapping,1,256*4*8,MAP_DUMP);
 fwrite(Write_Bank8Offset,1,256*4*8,MAP_DUMP);
 fwrite(Read_BankMapping,1,256*4,MAP_DUMP);
 fwrite(Write_BankMapping,1,256*4,MAP_DUMP);
 fclose(MAP_DUMP);
}
}

int ROM_has_header = Undetected;
/* Off = LoROM, On = HiROM, Undetected = autodetect */
int ROM_memory_map = Undetected;
int ROM_interleaved = Undetected;
int ROM_video_standard = Undetected;

void Load_LoROM(FILE *infile){
 // Read in as 32k blocks and (de)interleave
 for (int cnt = 0; cnt < NumBanks; cnt++)   // Read first half
  fread(RomAddress+cnt*65536+32768,1,32768,infile);

 for (int cnt = 0; cnt < NumBanks; cnt++)   // Read second half
  fread(RomAddress+cnt*65536,1,32768,infile);
}

unsigned check_for_header(FILE *fp, int filesize)
{
 unsigned ROM_start;

 fseek(fp, 0, SEEK_END);
 filesize = ftell(fp);

 // Easiest way to detect ROM header
  // Improvement suggested by Gridle (implemented 22/4/99)
  // check against even multiple of 1k instead of 32k
 // If ROM filesize is even multiple of 1k, assume no header
 // If ROM filesize is even multiple of 1k, +512, assume header
 if ((((filesize % 1024) == 0) || (ROM_has_header == Off)) &&
  (ROM_has_header != On)) ROM_start = 0;
 else if (((filesize % 1024) == ROM_Header_Size) || (ROM_has_header == On))
  ROM_start = ROM_Header_Size;
 else { // Basic header detection failure
/* Gridle 04/03/1998 (dmy)

ROM header in SMC files at 0x100 - 0x1FF usually is zero. I couldn't find
any other reliable way to check if the ROM is really SMC or BIN.

Originally I checked everything from 0x10 - 0x1FF, but one of my ROMs
(Brainies) had some stupid pirate group stuff just before 0x100, so I
changed this to search from 0x100 - 0x1FF.

This is now changed to 0x100 - 0x1F9 because of the king arthur ROM.

Charles Bilyue' 06/01/1999 (dmy)

(Check was later changed to 0x110 - 0x1F9 for some reason?)

LoROM checksums at 7FDC/7FDE, 81DC/81DE with header
HiROM checksums at FFDC/FFDE, 101DC/101DE with header

*/
  ROM_Header Header;     // For checking if the ROM has a header

  fseek(fp,0,SEEK_SET);
  fread(&Header,ROM_Header_Size,1,fp);

  ROM_start=ROM_Header_Size;

  for (unsigned cnt=0x110-11;cnt<0x1FA-11;cnt++)
   if(Header.RestOfHeader[cnt]!=0){ ROM_start=0; break; }
 }
/*
 if(ROM_start == ROM_Header_Size)
  printf("Header detected and ignored.\n");
 else
  printf("No header detected.\n");
 */
 return ROM_start;
}

bool LoadROM_normal(char *Filename){
 FILE *infile=fopen(Filename,"rb");
 if(!infile) return FALSE;  // File aint there m8

 unsigned ROM_start;        // This is where the ROM code itself starts.

 fseek(infile, 0, SEEK_END);
 int infilesize = ftell(infile);
 ROM_start = check_for_header(infile, infilesize);

 if(ROM_start == ROM_Header_Size)
  printf("Header detected and ignored.\n");
 else
  printf("No header detected.\n");

 ROM_format = Undetected;

 fseek(infile,0x7FC0+ROM_start,SEEK_SET);
 fread(&RomInfoLo,sizeof(SNESRomInfoStruct),1,infile);
 fseek(infile,0xFFC0+ROM_start,SEEK_SET);
 fread(&RomInfoHi,sizeof(SNESRomInfoStruct),1,infile);

 if((RomInfoLo.Checksum^RomInfoLo.Complement)==0xFFFF){
  if((RomInfoLo.ROM_makeup&0x0F)==1){
   strcpy(rom_romhilo,"Interleaved HiROM detected");
   ROM_format = HiROM_Interleaved;
  } else {
   strcpy(rom_romhilo,"LoROM detected");
   ROM_format = LoROM;
  }
 } else {
  if(((RomInfoHi.Checksum^RomInfoHi.Complement)==0xFFFF)){
   strcpy(rom_romhilo,"HiROM detected");
   ROM_format = HiROM;
  } else {
   strcpy(rom_romhilo,"Detection failed, using LoROM");
   ROM_format = LoROM;
  }
 }

 switch (ROM_interleaved){
  case On: ROM_format |= Interleaved; break;
  case Off: ROM_format &= ~Interleaved; break;
 }

 switch (ROM_memory_map){
  case HiROM:
   if (!(ROM_format & HiROM)) ROM_format = HiROM;
   strcpy(rom_romhilo,"HiROM forced"); break;
  case LoROM:
   if (ROM_format & HiROM) ROM_format = LoROM;
   strcpy(rom_romhilo,"LoROM forced"); break;
 }

 NumBanks = (((infilesize - ROM_start) + 65535) / 65536);
 if (NumBanks > 64) NumBanks = 64;  // Maximum 32Mbit ROM size for LoROM
 if (Allocate_ROM())   // Dynamic allocation of ROM
  return FALSE;        // return false if no memory left

 fseek(infile,ROM_start,SEEK_SET);

 switch(ROM_format){
  case HiROM:
   DisplayRomStats(&RomInfoHi);

   for (int cnt = 0; cnt < NumBanks; cnt++){    // Read in ROM
    fread(RomAddress+cnt*65536,1,65536,infile);
   }

   Set_HiROM_Map();
   break;

  case HiROM_Interleaved:
   DisplayRomStats(&RomInfoLo);
   Load_LoROM(infile);
   Set_HiROM_Map();
   break;

  case LoROM_Interleaved:
   printf("Interleaved LoROM not supported - basic LoROM loader used\n");
  case LoROM:
  default:
   DisplayRomStats(&RomInfoLo);
   Load_LoROM(infile);
   Set_LoROM_Map();
   break;
 }
 fclose(infile);

 return TRUE;
}

bool Load_LoROM_split(FILE *infile, char *Filename, int parts, long total_size){
 char tempname[MAXPATH];
 long bytes_read = 0;
 int part = 1;
 int infilesize;
 unsigned ROM_start;

 fnsplit(Filename, fn_drive, fn_dir, fn_file, fn_ext);

 // Read in as 32k blocks and interleave
 for (int cnt = 0;
      cnt < NumBanks && part <= parts; cnt++){ // Read first half
  infilesize = fread(RomAddress + cnt * 65536 + 32768, 1, 32768, infile);
  if (infilesize != EOF) bytes_read += infilesize;

  while ((bytes_read % 32768) || (infilesize < 1)){
  // partial bank read, must complete (or nothing read, go to next file)

   // next file
   fclose(infile);
   if (++part > parts){
    if (bytes_read != total_size) return FALSE;
    else break;
   }

   sprintf(fn_ext, ".%d", part);
   fnmerge(tempname, fn_drive, fn_dir, fn_file, fn_ext);

   FILE *infile=fopen(tempname,"rb");
   if (!infile) return FALSE;

   fseek(infile, 0, SEEK_END);
   infilesize = ftell(infile);
   ROM_start = check_for_header(infile, infilesize);
   total_size -= ROM_start;
   fseek(infile,ROM_start,SEEK_SET);

   infilesize =
    fread(RomAddress + cnt * 65536 + 32768 + (bytes_read % 32768), 1,
    32768 - (bytes_read % 32768), infile);

   if (infilesize != EOF) bytes_read += infilesize;
  }
 }

 for (int cnt = 0;
      cnt < NumBanks && part <= parts; cnt++){ // Read second half
  infilesize = fread(RomAddress + cnt * 65536, 1, 32768, infile);
  if (infilesize != EOF) bytes_read += infilesize;

  while ((bytes_read % 32768) || (infilesize < 1)){
  // partial bank read, must complete (or nothing read, go to next file)

   // next file
   fclose(infile);
   if (++part > parts){
    if (bytes_read != total_size) return FALSE;
    else break;
   }

   sprintf(fn_ext, ".%d", part);
   fnmerge(tempname, fn_drive, fn_dir, fn_file, fn_ext);

   FILE *infile=fopen(tempname,"rb");
   if (!infile) return FALSE;

   fseek(infile, 0, SEEK_END);
   infilesize = ftell(infile);
   ROM_start = check_for_header(infile, infilesize);
   total_size -= ROM_start;
   fseek(infile,ROM_start,SEEK_SET);

   infilesize =
    fread(RomAddress + cnt * 65536 + (bytes_read % 32768), 1,
    32768 - (bytes_read % 32768), infile);

   if (infilesize != EOF) bytes_read += infilesize;
  }
 }

 return TRUE;
}

bool LoadROM_split(char *Filename){
 char tempname[MAXPATH];
 int parts;
 long total_size = 0;

 fnsplit(Filename, fn_drive, fn_dir, fn_file, fn_ext);
 for (parts = 1; parts < 1000; parts++){
  sprintf(fn_ext, ".%d", parts + 1);
  fnmerge(tempname, fn_drive, fn_dir, fn_file, fn_ext);
  FILE *infile=fopen(tempname,"rb");
  if(!infile) break;
  fseek(infile, 0, SEEK_END);
  total_size += ftell(infile);
  fclose(infile);
 }
 if (parts == 1) return LoadROM_normal(Filename);
 printf("Split ROM image detected - %d parts found.\n", parts);

 FILE *infile=fopen(Filename,"rb");
 if(!infile) return FALSE;  // File aint there m8

 unsigned ROM_start;        // This is where the ROM code itself starts.

 fseek(infile, 0, SEEK_END);
 int infilesize = ftell(infile);
 ROM_start = check_for_header(infile, infilesize);
 total_size += infilesize - ROM_start;

 ROM_format = Undetected;

 unsigned long RomInfoHi_Read;

 fseek(infile,0x7FC0+ROM_start,SEEK_SET);
 if (fread(&RomInfoLo,sizeof(SNESRomInfoStruct),1,infile) != 1) return FALSE;
 fseek(infile,0xFFC0+ROM_start,SEEK_SET);
 RomInfoHi_Read = fread(&RomInfoHi,sizeof(SNESRomInfoStruct),1,infile);

 // if checksum and complement match, or HiROM info block couldn't be read
 if((RomInfoLo.Checksum^RomInfoLo.Complement)==0xFFFF ||
     RomInfoHi_Read != 1){
  if((RomInfoLo.ROM_makeup&0x0F)==1){
   strcpy(rom_romhilo,"Interleaved HiROM detected");
   ROM_format = HiROM_Interleaved;
  } else {
   strcpy(rom_romhilo,"LoROM detected");
   ROM_format = LoROM;
  }
 } else {
  if(((RomInfoHi.Checksum^RomInfoHi.Complement)==0xFFFF)){
   strcpy(rom_romhilo,"HiROM detected");
   ROM_format = HiROM;
  } else {
   strcpy(rom_romhilo,"Detection failed, using LoROM");
   ROM_format = LoROM;
  }
 }

 switch (ROM_memory_map){
  case HiROM:
   if (!(ROM_format & HiROM)) ROM_format = HiROM;
   strcpy(rom_romhilo,"HiROM forced"); break;
  case LoROM:
   if (ROM_format & HiROM) ROM_format = LoROM;
   strcpy(rom_romhilo,"LoROM forced"); break;
 }

 switch (ROM_interleaved){
  case On: ROM_format |= Interleaved; break;
  case Off: ROM_format &= ~Interleaved; break;
 }

 NumBanks = ((total_size + 65535) / 65536);
 if (NumBanks > 64) NumBanks = 64;  // Maximum 32Mbit ROM size for LoROM
 if (Allocate_ROM())   // Dynamic allocation of ROM
  return FALSE;        // return false if no memory left

 fseek(infile,ROM_start,SEEK_SET);

 switch(ROM_format){
  long bytes_read;

  case HiROM:

   DisplayRomStats(&RomInfoHi);

   bytes_read = fread(RomAddress,1,total_size,infile);
   fclose(infile);

   fnsplit(Filename, fn_drive, fn_dir, fn_file, fn_ext);
   for (int part = 2; part <= parts; part++){
    sprintf(fn_ext, ".%d", part);
    fnmerge(tempname, fn_drive, fn_dir, fn_file, fn_ext);

    FILE *infile=fopen(tempname,"rb");
    if (!infile) break;

    fseek(infile, 0, SEEK_END);
    infilesize = ftell(infile);
    ROM_start = check_for_header(infile, infilesize);
    total_size -= ROM_start;
    fseek(infile,ROM_start,SEEK_SET);

    infilesize =
     fread(RomAddress + bytes_read, 1, total_size - bytes_read, infile);
    fclose(infile);
    if (infilesize == EOF) break;
    if (total_size == bytes_read) break;
    bytes_read += infilesize;
   }
   if (infilesize == EOF) return FALSE;
   if (!infile) return FALSE;
   Set_HiROM_Map();
   break;
  case HiROM_Interleaved:
   DisplayRomStats(&RomInfoLo);
   Set_HiROM_Map();
   return Load_LoROM_split(infile, Filename, parts, total_size);

   break;
  case LoROM_Interleaved:
   printf("Split interleaved LoROM not supported - split basic LoROM loader used\n");
  case LoROM:
  default:
   DisplayRomStats(&RomInfoLo);
   Set_LoROM_Map();
   return Load_LoROM_split(infile, Filename, parts, total_size);

   break;
 }

 return TRUE;
}

int LoadROM(char *Filename){
 ROMFileType filetype;

 SaveSRAM(SRAM_filename);   // Ensures SRAM saved before loading new ROM
 ROM_Loaded=FALSE;
 SaveRamLength=0;
 if(ROM_filename){ free(ROM_filename); ROM_filename=0; }

 filetype = GetROMFileType(Filename);

 ROM_filename=(char *) malloc(strlen(Filename));
 if(ROM_filename) strcpy(ROM_filename,Filename);
 if(!rom_romfile) rom_romfile=(char *)malloc(100);
 if(!rom_romhilo) rom_romhilo=(char *)malloc(100);
 if(!rom_romtype) rom_romtype=(char *)malloc(50);
 if(!rom_romsize) rom_romsize=(char *)malloc(50);
 if(!rom_romname) rom_romname=(char *)malloc(50);
 if(!rom_sram)    rom_sram   =(char *)malloc(50);
 if(!rom_country) rom_country=(char *)malloc(50);

 strcpy(rom_romfile,"File name: ");
 strcat(rom_romfile,Filename);
 rom_romfile[33]=0;

 switch (filetype)
 {
  case ROMFileType_split: ROM_Loaded = LoadROM_split(Filename); break;
  case ROMFileType_normal:
  default: ROM_Loaded = LoadROM_normal(Filename);
 }

 if (ROM_Loaded == FALSE) return FALSE;
 Reset_Memory();
 Reset_SRAM();
 Reset_CPU();
 Reset_SPC();
 Reset_Sound();

 LoadSRAM(SRAM_filename);     // This loads in the Save RAM

 printf("RES: %04lX\n"
        "NMI N,E: %04lX,%04lX\n"
        "IRQ N,E: %04lX,%04lX\n"
        "BRK N,E: %04lX,%04lX\n"
        "COP N,E: %04lX,%04lX\n",
  _65816_PC & 0xFFFF,
  NMI_Nvector, NMI_Evector,
  IRQ_Nvector, IRQ_Evector,
  BRK_Nvector, IRQ_Evector,
  COP_Nvector, COP_Evector);

 return TRUE;
}

char *TypeTable[]={
 "ROM",
 "ROM+RAM",
 "ROM+SRAM",
 "ROM+DSP1",
 "ROM+RAM+DSP1",
 "ROM+SRAM+DSP1",
 "Unknown",
 "Unknown FX chip",
 "FX-V1",
 "FX(Argonaut)",
 "FX-V2",
 "FX(SA-1)",
 "FX(?KSS?)",
 "FX(Capcom)",
 "FX(S-DD1)",
 "FX(PLGS)",
 "FX(Gameboy)",
 "FX(BS-X)",
 "FX(Capcom C4)",
 "FX(SETA) (DSP2?)"
};

char *CountryTable[]={
 "Japan",
 "USA",
 "Europe, Oceania, Asia",
 "Sweden",
 "Finland",
 "Denmark",
 "France",
 "Holland",
 "Spain",
 "Germany, Austria, Switz.",
 "Italy",
 "Hong Kong, China",
 "Indonesia",
 "Korea"
};

extern "C" unsigned FPSMaxTicks;
void DisplayRomStats(SNESRomInfoStruct *RomInfo){
 char TempBuffer[30];

 // ROM Type

 strcpy(rom_romtype,"ROM Type: ");
 switch(RomInfo->ROM_type){
  case 0:   // ROM
  case 1:   // ROM/RAM
  case 2:   // ROM/SRAM
  case 3:   // ROM/DSP1
  case 4:   // ROM/RAM/DSP1
  case 5:   // ROM/SRAM/DSP1
   strcat(rom_romtype,TypeTable[RomInfo->ROM_type]); break;
  case 0x10 ... 0x12:   case 0x16:
  case 0x20 ... 0x26:   case 0x30 ... 0x34:
  case 0x40 ... 0x42:   case 0x44:  case 0x46:
  case 0x50 ... 0x54:   case 0x56:
  case 0x60 ... 0x66:   case 0x70 ... 0x76:
  case 0x80 ... 0x86:   case 0x90 ... 0x96:
  case 0xA0 ... 0xA6:   case 0xB0 ... 0xB6:
  case 0xC0 ... 0xC6:   case 0xD0 ... 0xD6:
  case 0xE0 ... 0xE2:   case 0xE4:  case 0xE6:
  case 0xF0 ... 0xF2:   case 0xF4 ... 0xF5:
            // Unknown FX chip
   strcat(rom_romtype,TypeTable[7]);
  case 0x13:    // FX-V1
   strcat(rom_romtype,TypeTable[8]); break;
  case 0x14:    // FX(Argonaut)
   strcat(rom_romtype,TypeTable[9]); break;
  case 0x15:    // FX-V2
   strcat(rom_romtype,TypeTable[10]); break;
  case 0x35:    // FX(SA-1)
   strcat(rom_romtype,TypeTable[11]); break;
  case 0x36:    // FX(?KSS?)
   strcat(rom_romtype,TypeTable[12]); break;
  case 0x43:    // FX(Capcom)
   strcat(rom_romtype,TypeTable[13]); break;
  case 0x45:    // FX(S-DD1)
   strcat(rom_romtype,TypeTable[14]); break;
  case 0x55:    // FX(PLGS)
   strcat(rom_romtype,TypeTable[15]); break;
  case 0xE3:    // FX(Gameboy)
   strcat(rom_romtype,TypeTable[16]); break;
  case 0xE5:    // FX(BS-X)
   strcat(rom_romtype,TypeTable[17]); break;
  case 0xF3:    // FX(Capcom C4)
   strcat(rom_romtype,TypeTable[18]); break;
  case 0xF6:    // FX(SETA) (DSP2?)
   strcat(rom_romtype,TypeTable[19]); break;
  default:      // Unknown
   strcat(rom_romtype,TypeTable[6]);
 }

 // ROM Size

 strcpy(rom_romsize,"ROM Size : ");
 strcat(rom_romsize,itoa(1 << ((RomInfo->ROM_size)-7),TempBuffer,10));
 strcat(rom_romsize,"Mbits");

 // ROM SRAM size

 SaveRamLength=RomInfo->SRAM_size;
 if(((RomInfo->ROM_type)==2) || ((RomInfo->ROM_type)>=5) &&
  SaveRamLength){
  SaveRamLength=1024 << SaveRamLength;

  strcpy(rom_sram,"SRAM Size: ");
  strcat(rom_sram,itoa(SaveRamLength*8/1024,TempBuffer,10));
  strcat(rom_sram,"Kbits");
  if(SaveRamLength>65536) SaveRamLength=65536;
 } else {
  SaveRamLength=0;
  strcpy(rom_sram,"SRAM Size: No SRAM");
 }

 // ROM Country

 if (RomInfo->Country_code <= 13)
 {
  sprintf(rom_country, "Country: %s", CountryTable[ RomInfo->Country_code ]);
  if (((RomInfo->Country_code < 2) && (ROM_video_standard != PAL_video))
   || (ROM_video_standard == NTSC_video))
  {
   strcat(rom_country,
    (ROM_video_standard != NTSC_video) ? " (NTSC)" : " (NTSC forced)");

   SNES_COUNTRY = 0x00;

   Frametime = BPS_TO_TIMER(60);
   FPSMaxTicks = 60;
   LastVBLLine = 261;
  }
  else if (((RomInfo->Country_code < 14) && (ROM_video_standard != NTSC_video))
   || (ROM_video_standard == PAL_video))
  {
   strcat(rom_country,
    (ROM_video_standard != PAL_video) ? " (PAL)" : " (PAL forced)");

   SNES_COUNTRY = 0x10;

   Frametime = BPS_TO_TIMER(50);
   FPSMaxTicks = 50;
   LastVBLLine = 311;
  }
 } else {
  switch (ROM_video_standard)
  {
   case PAL_video:
    strcpy(rom_country, "Country: Unknown (PAL forced)");

    SNES_COUNTRY = 0x10;

    Frametime = BPS_TO_TIMER(50);
    FPSMaxTicks = 50;
    LastVBLLine = 311;
    break;

   case NTSC_video:
    strcpy(rom_country, "Country: Unknown (NTSC forced)");

    SNES_COUNTRY = 0x00;

    Frametime = BPS_TO_TIMER(60);
    FPSMaxTicks = 60;
    LastVBLLine = 261;
    break;

   default:
    strcpy(rom_country, "Country: Unknown (Using NTSC)");

    SNES_COUNTRY = 0x00;

    Frametime = BPS_TO_TIMER(60);
    FPSMaxTicks = 60;
    LastVBLLine = 261;
  }
 }

 // ROM Name

 for(int i = 0; i < 21; i++)
  TempBuffer[i] = RomInfo->ROM_Title[i];

 TempBuffer[21] = 0;

 strcpy(rom_romname,"ROM Title: ");
 strcat(rom_romname,TempBuffer);

 // Print information

 printf("\n%s\n",rom_romname);
 printf("%s\n",rom_romtype);
 if((RomInfo->ROM_type)>2)
  printf("Some extra hardware used by this ROM is not yet supported!\n");
 printf("%s\n",rom_romhilo);
 printf("%s\n",rom_romsize);
 printf("%s\n",rom_sram);
 printf("%s\n",rom_country);

 // Ugly things, but helps with 256x240 in ROM information window
 rom_romhilo[33]=0;
 rom_country[33]=0;
}

void Wangle(void){
 if(ROM_Loaded){
#ifdef DEBUG
#ifdef HDMATRACKER
  FILE *HDMA_DUMP = fopen("HDMA.DMP","wb");
  if (HDMA_DUMP)
  {
   fwrite(HDMAAddress+LastHDMA,1,1024-LastHDMA,HDMA_DUMP);
   fwrite(HDMAAddress,1,LastHDMA,HDMA_DUMP);
   fclose(HDMA_DUMP);
  }
#endif
#ifdef DMATRACKER
  FILE *DMA_DUMP = fopen("DMA.DMP","wb");
  if (DMA_DUMP)
  {
   fwrite(DMAAddress+LastDMA,1,1024-LastDMA,DMA_DUMP);
   fwrite(DMAAddress,1,LastDMA,DMA_DUMP);
   fclose(DMA_DUMP);
  }
#endif
#ifdef SPCTRACKER
  // This saves the SPC-tracker dump!
  FILE *SPC_DUMP = fopen("SPC.DMP","wb");
  if (SPC_DUMP)
  {
   fwrite(SPC_InsAddress+SPC_LastIns,1,(SPCTRACKER)-SPC_LastIns,SPC_DUMP);
   fwrite(SPC_InsAddress,1,SPC_LastIns,SPC_DUMP);
   fclose(SPC_DUMP);
  }
#endif
#ifdef CPUTRACKER
  // This saves the CPU-tracker dump!
  FILE *INS_DUMP = fopen("C:\\INS.DMP","wb");
  if (INS_DUMP)
  {
#if CPUTRACKER >= 16
   fwrite(InsAddress+LastIns,1,(CPUTRACKER)-LastIns,INS_DUMP);
   fwrite(InsAddress,1,LastIns,INS_DUMP);
#else
   fwrite(InsAddress,1,5120,INS_DUMP);
#endif
   fclose(INS_DUMP);
  }
#endif
  FILE *SPCRAM_DUMP = fopen("SPCRAM.DMP","wb");
  if (SPCRAM_DUMP)
  {
   fwrite(SPCRAM,1,65536,SPCRAM_DUMP);
   fclose(SPCRAM_DUMP);
  }

   FILE *WRAM_DUMP = fopen("WRAM.DMP","wb");
  if (WRAM_DUMP)
  {
   fwrite(WRAM,1,131072,WRAM_DUMP);
   fclose(WRAM_DUMP);
  }

  FILE *VRAM_DUMP = fopen("VRAM.DMP","wb");
  if (VRAM_DUMP)
  {
   fwrite(VRAM,1,65536,VRAM_DUMP);
   fclose(VRAM_DUMP);
  }

#endif
 }
}
