; SNEeSe SPC700 CPU emulation core
; Originally written by Lee Hammerton in AT&T assembly
; Maintained/rewritten/ported to NASM by Charles Bilyue'
;
; Compile under NASM

;%define TRACKERS 1048576

extern _SPCRAM
extern _SPC_CPU_cycle_multiplicand,_SPC_CPU_cycle_divisor
extern SPC_CPU_cycles_mul,SPC_CPU_cycles
extern _TotalCycles,_CycleLatch0,_CycleLatch1,_CycleLatch2
extern Read_Map_21,Write_Map_21
extern _SPC_CTRL,_SPC_DSP
extern _SPC_DSP_ADDR,_SPC_DSP_DATA
extern _SPC_READ_DSP,_SPC_WRITE_DSP
extern _SPC_CNT0,_SPC_CNT1,_SPC_CNT2
extern _SPC_TIMT0,_SPC_TIMT1,_SPC_TIMT2
extern _SPC_TIMER0,_SPC_TIMER1,_SPC_TIMER2
extern _Update_SPC_Timer_0,_Update_SPC_Timer_1,_Update_SPC_Timer_2
extern _Wrap_SPC_Cyclecounter

extern SNES_Cycles,SPC_last_cycles,CPU_Executing
extern _Map_Byte,_Map_Address
extern _InvalidSPCOpcode,_DisplaySPC

; This file contains:
;  CPU core info
;  Reset
;  Execution Loop
;  Invalid Opcode Handler
;  Flag format conversion tables
;  Variable definitions (registers, interrupt vectors, etc.)
;  CPU opcode emulation handlers
;  CPU opcode handler table
;  CPU opcode timing table
;
; CPU core info:
;  Nearly all general registers are now used in SPC700 emulation:
;   EAX,EBX are used by the memory mapper;
;   CH is used to hold PSW register;
;   EDX is used as CPU work register;
;   EBP is used to hold cycle counter;
;   ESI is used by the opcode fetcher;
;   EDI is used as memory mapper work register.
;
;    A register              - _A
;    Y register              - _Y
;    YA register pair        - _YA
;    X register              - _X
;    Stack pointer           - _SP
;    Program Counter         - _PC
;    Processor status word   - _PSW
;       True x86 layout = |V|-|-|-|S|Z|-|A|-|-|-|C|
;    True SPC700 layout =         |N|V|P|B|H|I|Z|C|
;                   Using         |N|Z|P|H|B|I|V|C|
;
; SPC timers
;  SPC700 timing is directly related to 65c816 timing, SPC gets 3 cycles for
;  every 32 (21.48 MHz) 65c816 cycles. Since the timers run at 8KHz and 64KHz
;  and the main chip runs at 2.048Mhz, the timers are clocked as follows:
;    2.048MHz / 8KHz  = 256 cycles    (Timers 0 and 1)
;    2.048MHz / 64KHz = 32  cycles    (Timer 2)
;
;

%include "misc.ni"
%include "SPC.ni"

EXPORT_C SPC_start

; These are the bits for flag set/clr operations
SPC_FLAG_C equ 1    ; Carry
SPC_FLAG_V equ 2    ; Overflow
SPC_FLAG_I equ 4    ; Interrupt Disable
SPC_FLAG_B equ 8    ; Break
SPC_FLAG_H equ 0x10 ; Half-carry
SPC_FLAG_P equ 0x20 ; Page (direct page)
SPC_FLAG_Z equ 0x40 ; Zero result
SPC_FLAG_N equ 0x80 ; Negative result

SPC_FLAG_NZ equ (SPC_FLAG_N | SPC_FLAG_Z)
SPC_FLAG_NZC equ (SPC_FLAG_NZ | SPC_FLAG_C)
SPC_FLAG_NHZC equ (SPC_FLAG_NZC | SPC_FLAG_H)

REAL_SPC_FLAG_C equ 1       ; Carry
REAL_SPC_FLAG_Z equ 2       ; Zero result
REAL_SPC_FLAG_I equ 4       ; Interrupt Disable
REAL_SPC_FLAG_H equ 8       ; Half-carry
REAL_SPC_FLAG_B equ 0x10    ; Break
REAL_SPC_FLAG_P equ 0x20    ; Page (direct page)
REAL_SPC_FLAG_V equ 0x40    ; Overflow
REAL_SPC_FLAG_N equ 0x80    ; Negative result

%define _PSW __SPC_PSW
%define _YA __SPC_YA
%define _A  __SPC_A
%define _Y  __SPC_Y
%define _X  __SPC_X
%define _SP __SPC_SP
%define _PC __SPC_PC

%define R_PSW_B      R_SPC700_PSW_B
%define R_Base       R_SPC700_Base
%define R_Cycles     R_SPC700_Cycles
%define R_NativePC   R_SPC700_NativePC

%define B_SPC_Code_Base     [R_Base-SPC_Register_Base+SPC_Code_Base]
%define B_PC                [R_Base-SPC_Register_Base+_PC]
%define B_YA                [R_Base-SPC_Register_Base+_YA]
%define B_A                 [R_Base-SPC_Register_Base+_A]
%define B_Y                 [R_Base-SPC_Register_Base+_Y]
%define B_SPC_PAGE          [R_Base-SPC_Register_Base+SPC_PAGE]
%define B_SP                [R_Base-SPC_Register_Base+_SP]
%define B_PSW               [R_Base-SPC_Register_Base+_PSW]
%define B_X                 [R_Base-SPC_Register_Base+_X]
%define B_SPC_PORT0R        [R_Base-SPC_Register_Base+_SPC_PORT0R]
%define B_SPC_PORT1R        [R_Base-SPC_Register_Base+_SPC_PORT1R]
%define B_SPC_PORT2R        [R_Base-SPC_Register_Base+_SPC_PORT2R]
%define B_SPC_PORT3R        [R_Base-SPC_Register_Base+_SPC_PORT3R]
%define B_SPC_PORT0W        [R_Base-SPC_Register_Base+_SPC_PORT0W]
%define B_SPC_PORT1W        [R_Base-SPC_Register_Base+_SPC_PORT1W]
%define B_SPC_PORT2W        [R_Base-SPC_Register_Base+_SPC_PORT2W]
%define B_SPC_PORT3W        [R_Base-SPC_Register_Base+_SPC_PORT3W]
%ifdef DEBUG
%define B_SPC_TEMP_ADD      [R_Base-SPC_Register_Base+SPC_TEMP_ADD]
%endif

; Load PSW to register R_PSW_B
%macro LOAD_PSW 0
 mov byte R_PSW_B,[_PSW]
%endmacro

; Save register R_PSW_B to PSW
%macro SAVE_PSW 0
 mov byte [_PSW],R_PSW_B
%endmacro

; Load cycle counter to register R_Cycles
%macro LOAD_CYCLES 0
  mov dword R_Cycles,[_TotalCycles]
%endmacro

; Save register R_Cycles to cycle counter
%macro SAVE_CYCLES 0
  mov dword [_TotalCycles],R_Cycles
%endmacro

; Load base pointer to CPU register set
%macro LOAD_BASE 0
  mov R_Base,SPC_Register_Base
%endmacro

; Load register R_NativePC with pointer to code at PC
%macro LOAD_PC 0
 mov dword R_NativePC,[SPC_Code_Base]
 add dword R_NativePC,[_PC]
%endmacro

; Get PC from register R_NativePC
;%1 = with
%macro GET_PC 1
 mov %1,R_NativePC
 sub dword %1,[SPC_Code_Base]
%endmacro

; Save PC from register R_NativePC
;%1 = with
%macro SAVE_PC 1
 GET_PC %1
 mov dword [_PC],%1
%endmacro

; Set up the flags from PC flag format to SPC flag format stored in %al
; Corrupts eax
;%1 = from, %2 = to
%macro SETUPFLAGS_SPC 0-2 R_PSW_B,al
;%macro Flags_Native_to_SPC 0-2 R_PSW_B,al
 xor eax,eax
 mov al,%1
 mov byte %2,[_SPC_Setup_Flags_Table+eax]
%endmacro

; Restore flags from SPC format in %al to PC format
; Corrupts eax
;%1 = to
%macro RESTOREFLAGS_SPC 0-1 R_PSW_B
;%macro Flags_SPC_to_Native 0-1 R_PSW_B
 and eax,0xFF
 test al,REAL_SPC_FLAG_P
 mov al,[_SPC_Restore_Flags_Table+eax]
 setnz [SPC_PAGE+1] ; Setup direct page bit
 mov %1,al
%endmacro

; SPC MEMORY MAPPER IS PLACED HERE (ITS SIMPLER THAN THE CPU ONE!)

; bx - contains the actual address, al is where the info should be stored, edi is free
; NB bx is not corrupted! edi is corrupted!
; NB eax is not corrupted barring returnvalue in al... e.g. ah should not be used etc!

ALIGNC
SPC_READ_MAPPER:
 and ebx,0xFFFF
 test bh,bh
 jz SPC_READ_ZERO_PAGE

 cmp ebx,0xFFC0
 jae SPC_READ_RAM_ROM
SPC_READ_RAM:
 mov al,[_SPCRAM+ebx]
 ret

ALIGNC
SPC_READ_RAM_ROM:
 mov al,[SPC_ROM_CODE - 0xFFC0 + ebx]
SPC_FFC0_Address equ $-4
 ret

ALIGNC
SPC_READ_ZERO_PAGE:
 cmp bl,0xF0
 jb SPC_READ_RAM

SPC_READ_FUNC:
 jmp dword [Read_Func_Map - 0xF0 * 4 + ebx * 4]

ALIGNC
SPC_READ_INVALID:
 mov al,0xFF    ; v0.15
%ifdef DEBUG
extern _InvalidSPCHWRead
;and ebx,0xFFFF
 mov [_Map_Address],ebx ; Set up Map Address so message works!
 mov [_Map_Byte],al     ; Set up Map Byte so message works

 pusha
 call _InvalidSPCHWRead ; Display read from invalid HW warning
 popa
%endif
 ret

;   --------

SPC_WRITE_MAPPER:
 and ebx,0xFFFF
 test bh,bh
 jz SPC_WRITE_ZERO_PAGE

SPC_WRITE_RAM:
 mov [_SPCRAM+ebx],al
 ret

ALIGNC
SPC_WRITE_ZERO_PAGE:
 cmp bl,0xF0
 jb SPC_WRITE_RAM

SPC_WRITE_FUNC:
 jmp dword [Write_Func_Map - 0xF0 * 4 + ebx * 4]

SPC_WRITE_INVALID:
%ifdef DEBUG
extern _InvalidSPCHWWrite
;and ebx,0xFFFF
 mov [_Map_Address],ebx     ; Set up Map Address so message works!
 mov [_Map_Byte],al         ; Set up Map Byte so message works

 pusha
 call _InvalidSPCHWWrite    ; Display write to invalid HW warning
 popa
%endif
 ret

; GET_BYTE & GET_WORD now assume ebx contains the read address and 
; eax the place to store value also, corrupts edi

%macro GET_BYTE_SPC 0
 call SPC_READ_MAPPER
%endmacro

%macro GET_WORD_SPC 0
 GET_BYTE_SPC
 mov ah,al
 inc bx
 GET_BYTE_SPC
 xchg ah,al
%endmacro

; SET_BYTE & SET_WORD now assume ebx contains the write address and 
; eax the value to write, corrupts edi

%macro SET_BYTE_SPC 0
 call SPC_WRITE_MAPPER
%endmacro

%macro  SET_WORD_SPC 0
 SET_BYTE_SPC
 mov al,ah
 inc bx
 SET_BYTE_SPC
%endmacro

; --- Ease up on the finger cramps ;-)

;%1 = flag
%macro SET_FLAG_SPC 1
 or R_PSW_B,%1
%endmacro

;%1 = flag
%macro CLR_FLAG_SPC 1
 and R_PSW_B,~(%1)
%endmacro

;%1 = flag
%macro CPL_FLAG_SPC 1
 xor R_PSW_B,%1
%endmacro

;%1 = flag
%macro TEST_FLAG_SPC 1  ; Set ZF=!flag
 test R_PSW_B,%1
%endmacro

;%1 = flag, %2 = wheretogo
%macro JUMP_FLAG_SPC 2
 TEST_FLAG_SPC %1
 jnz %2
%endmacro

;%1 = flag, %2 = wheretogo
%macro JUMP_NOT_FLAG_SPC 2
 TEST_FLAG_SPC %1
 jz %2
%endmacro

; Push / Pop Macros assume eax contains value - corrupt ebx,edi
%macro PUSH_B 0         ; Push Byte (SP--)
 mov ebx,B_SP
 mov [_SPCRAM+ebx],al   ; Store data on stack
 dec ebx
 mov B_SP,bl            ; Decrement S (Byte)
%endmacro

%macro POP_B 0          ; Pop Byte (++SP)
 mov ebx,B_SP
 inc bl
 mov B_SP,bl
 mov al,[_SPCRAM+ebx]   ; Fetch data from stack
%endmacro

%macro PUSH_W 0         ; Push Word (SP--)
 mov ebx,B_SP
 mov [_SPCRAM+ebx],ah   ; Store data on stack
 mov [_SPCRAM+ebx-1],al ; Store data on stack
 sub bl,2
 mov B_SP,bl            ; Postdecrement SP
%endmacro

%macro POP_W 0          ; Pop Word (++SP)
 mov ebx,B_SP
 add bl,2               ; Preincrement SP
 mov B_SP,bl
 mov ah,[_SPCRAM+ebx]   ; Fetch data from stack
 mov al,[_SPCRAM+ebx-1] ; Fetch data from stack
%endmacro

; Set SPC700 flags to PC flags
; Can only set flags NHZC - corrupts ah
;%1 = flags
%macro STORE_FLAGS_SPC 1
 lahf
 STORE_FLAGS_SPC_FROM %1,ah
%endmacro

; Set SPC700 flags to PC flags
; Can only set flags NH11ZC - corrupts loc
; Assumes loc has PC flags (as from lahf)
;%1 = flags, %2 = loc
%macro STORE_FLAGS_SPC_FROM 2
 and R_PSW_B,(~%1)
 and %2,(%1)
 or R_PSW_B,%2
%endmacro

; Set PC Flags To SNES Flags (only sets required flags all others cleared!) corrupts eax

;%1 = flags
%macro MAKE_FLAGS_SPC 1
 mov ah,R_PSW_B
 and ah,%1      ; Flags ready
 sahf
%endmacro

; Set PC Flags To SNES Flags (Inverts bit (e.g. C=1 means clear carry)) corrupts eax

;%1 = flags
%macro MAKE_FLAGS_SPC_INV 1
 mov ah,R_PSW_B
 xor ah,0xFF
 and ah,%1      ; Flags ready
 sahf
%endmacro

; PC carry flag = SPC700 carry flag (all others undefined)
%macro MAKE_CARRY 0
 rol R_PSW_B,8
%endmacro

; PC carry flag = NOT SPC700 carry flag (all others undefined)
%macro MAKE_NOT_CARRY 0
 rol R_PSW_B,8
 cmc
%endmacro

section .data
ALIGND
EXPORT SPCOpTable
dd SPC_NOP          ,SPC_TCALL_0       ,SPC_SET1          ,SPC_BBS                ; 00
dd SPC_OR_A_dp      ,SPC_OR_A_abs      ,SPC_OR_A_OXO      ,SPC_OR_A_OOdp_XOO
dd SPC_OR_A_IM      ,SPC_OR_dp_dp      ,SPC_OR1           ,SPC_ASL_dp
dd SPC_ASL_abs      ,SPC_PUSH_PSW      ,SPC_TSET1         ,SPC_INVALID
dd SPC_BPL          ,SPC_TCALL_1       ,SPC_CLR1          ,SPC_BBC                ; 10
dd SPC_OR_A_Odp_XO  ,SPC_OR_A_Oabs_XO  ,SPC_OR_A_Oabs_YO  ,SPC_OR_A_OOdpO_YO
dd SPC_OR_dp_IM     ,SPC_OR_OXO_OYO    ,SPC_DECW_dp       ,SPC_ASL_Odp_XO
dd SPC_ASL_A        ,SPC_DEC_X         ,SPC_CMP_X_abs     ,SPC_JMP_Oabs_XO
dd SPC_CLRP         ,SPC_TCALL_2       ,SPC_SET1          ,SPC_BBS                ; 20
dd SPC_AND_A_dp     ,SPC_AND_A_abs     ,SPC_AND_A_OXO     ,SPC_AND_A_OOdp_XOO
dd SPC_AND_A_IM     ,SPC_AND_dp_dp     ,SPC_OR1C          ,SPC_ROL_dp
dd SPC_ROL_abs      ,SPC_PUSH_A        ,SPC_CBNE_dp       ,SPC_BRA
dd SPC_BMI          ,SPC_TCALL_3       ,SPC_CLR1          ,SPC_BBC                ; 30
dd SPC_AND_A_Odp_XO ,SPC_AND_A_Oabs_XO ,SPC_AND_A_Oabs_YO ,SPC_AND_A_OOdpO_YO
dd SPC_AND_dp_IM    ,SPC_AND_OXO_OYO   ,SPC_INCW_dp       ,SPC_ROL_Odp_XO
dd SPC_ROL_A        ,SPC_INC_X         ,SPC_CMP_X_dp      ,SPC_CALL
dd SPC_SETP         ,SPC_TCALL_4       ,SPC_SET1          ,SPC_BBS                ; 40
dd SPC_EOR_A_dp     ,SPC_EOR_A_abs     ,SPC_EOR_A_OXO     ,SPC_EOR_A_OOdp_XOO
dd SPC_EOR_A_IM     ,SPC_EOR_dp_dp     ,SPC_AND1          ,SPC_LSR_dp
dd SPC_LSR_abs      ,SPC_PUSH_X        ,SPC_TCLR1         ,SPC_PCALL
dd SPC_BVC          ,SPC_TCALL_5       ,SPC_CLR1          ,SPC_BBC                ; 50
dd SPC_EOR_A_Odp_XO ,SPC_EOR_A_Oabs_XO ,SPC_EOR_A_Oabs_YO ,SPC_EOR_A_OOdpO_YO
dd SPC_EOR_dp_IM    ,SPC_EOR_OXO_OYO   ,SPC_CMPW_YA_dp    ,SPC_LSR_Odp_XO
dd SPC_LSR_A        ,SPC_MOV_X__A      ,SPC_CMP_Y_abs     ,SPC_JMP_abs
dd SPC_CLRC         ,SPC_TCALL_6       ,SPC_SET1          ,SPC_BBS                ; 60
dd SPC_CMP_A_dp     ,SPC_CMP_A_abs     ,SPC_CMP_A_OXO     ,SPC_CMP_A_OOdp_XOO
dd SPC_CMP_A_IM     ,SPC_CMP_dp_dp     ,SPC_AND1C         ,SPC_ROR_dp
dd SPC_ROR_abs      ,SPC_PUSH_Y        ,SPC_DBNZ_dp       ,SPC_RET
dd SPC_BVS          ,SPC_TCALL_7       ,SPC_CLR1          ,SPC_BBC                ; 70
dd SPC_CMP_A_Odp_XO ,SPC_CMP_A_Oabs_XO ,SPC_CMP_A_Oabs_YO ,SPC_CMP_A_OOdpO_YO
dd SPC_CMP_dp_IM    ,SPC_CMP_OXO_OYO   ,SPC_ADDW_YA_dp    ,SPC_ROR_Odp_XO
dd SPC_ROR_A        ,SPC_MOV_A__X      ,SPC_CMP_Y_dp      ,SPC_INVALID
dd SPC_SETC         ,SPC_TCALL_8       ,SPC_SET1          ,SPC_BBS                ; 80
dd SPC_ADC_A_dp     ,SPC_ADC_A_abs     ,SPC_ADC_A_OXO     ,SPC_ADC_A_OOdp_XOO
dd SPC_ADC_A_IM     ,SPC_ADC_dp_dp     ,SPC_EOR1          ,SPC_DEC_dp
dd SPC_DEC_abs      ,SPC_MOV_Y_IM      ,SPC_POP_PSW       ,SPC_MOV_dp_IM
dd SPC_BCC          ,SPC_TCALL_9       ,SPC_CLR1          ,SPC_BBC                ; 90
dd SPC_ADC_A_Odp_XO ,SPC_ADC_A_Oabs_XO ,SPC_ADC_A_Oabs_YO ,SPC_ADC_A_OOdpO_YO
dd SPC_ADC_dp_IM    ,SPC_ADC_OXO_OYO   ,SPC_SUBW_YA_dp    ,SPC_DEC_Odp_XO
dd SPC_DEC_A        ,SPC_MOV_X__SP     ,SPC_DIV           ,SPC_XCN
dd SPC_EI           ,SPC_TCALL_10      ,SPC_SET1          ,SPC_BBS                ; A0
dd SPC_SBC_A_dp     ,SPC_SBC_A_abs     ,SPC_SBC_A_OXO     ,SPC_SBC_A_OOdp_XOO
dd SPC_SBC_A_IM     ,SPC_SBC_dp_dp     ,SPC_MOV1_C_       ,SPC_INC_dp
dd SPC_INC_abs      ,SPC_CMP_Y_IM      ,SPC_POP_A         ,SPC_MOV_OXOInc_A
dd SPC_BCS          ,SPC_TCALL_11      ,SPC_CLR1          ,SPC_BBC                ; B0
dd SPC_SBC_A_Odp_XO ,SPC_SBC_A_Oabs_XO ,SPC_SBC_A_Oabs_YO ,SPC_SBC_A_OOdpO_YO
dd SPC_SBC_dp_IM    ,SPC_SBC_OXO_OYO   ,SPC_MOV_YA_dp     ,SPC_INC_Odp_XO
dd SPC_INC_A        ,SPC_MOV_SP_X      ,SPC_INVALID       ,SPC_MOV_A_OXOInc
dd SPC_DI           ,SPC_TCALL_12      ,SPC_SET1          ,SPC_BBS                ; C0
dd SPC_MOV_dp__A    ,SPC_MOV_abs__A    ,SPC_MOV_OXO__A    ,SPC_MOV_OOdp_XOO__A
dd SPC_CMP_X_IM     ,SPC_MOV_abs__X    ,SPC_MOV1__C       ,SPC_MOV_dp__Y
dd SPC_MOV_abs__Y   ,SPC_MOV_X_IM      ,SPC_POP_X         ,SPC_MUL
dd SPC_BNE          ,SPC_TCALL_13      ,SPC_CLR1          ,SPC_BBC                ; D0
dd SPC_MOV_Odp_XO__A,SPC_MOV_Oabs_XO__A,SPC_MOV_Oabs_YO__A,SPC_MOV_OOdpO_YO__A
dd SPC_MOV_dp__X    ,SPC_MOV_Odp_YO__X ,SPC_MOV_dp_YA     ,SPC_MOV_Odp_XO__Y
dd SPC_DEC_Y        ,SPC_MOV_A__Y      ,SPC_CBNE_Odp_XO   ,SPC_INVALID
dd SPC_CLRV         ,SPC_TCALL_14      ,SPC_SET1          ,SPC_BBS                ; E0
dd SPC_MOV_A_dp     ,SPC_MOV_A_abs     ,SPC_MOV_A_OXO     ,SPC_MOV_A_OOdp_XOO
dd SPC_MOV_A_IM     ,SPC_MOV_X_abs     ,SPC_NOT1          ,SPC_MOV_Y_dp
dd SPC_MOV_Y_abs    ,SPC_NOTC          ,SPC_POP_Y         ,SPC_INVALID ;SPC_SLEEP
dd SPC_BEQ          ,SPC_TCALL_15      ,SPC_CLR1          ,SPC_BBC                ; F0
dd SPC_MOV_A_Odp_XO ,SPC_MOV_A_Oabs_XO ,SPC_MOV_A_Oabs_YO ,SPC_MOV_A_OOdpO_YO
dd SPC_MOV_X_dp     ,SPC_MOV_X_Odp_YO  ,SPC_MOV_dp_dp     ,SPC_MOV_Y_Odp_XO
dd SPC_INC_Y        ,SPC_MOV_Y__A      ,SPC_DBNZ_Y        ,SPC_INVALID ;SPC_STOP

; This holds the base instruction timings in cycles
SPCCycleTable:
db 2,8,4,5,3,4,3,6,2,6,5,4,5,4,6,8  ; 00
db 2,8,4,5,4,5,5,6,5,5,6,5,2,2,4,6  ; 10
db 2,8,4,5,3,4,3,6,2,6,5,4,5,4,5,4  ; 20
db 2,8,4,5,4,5,5,6,5,5,6,5,2,2,3,8  ; 30
db 2,8,4,5,3,4,3,6,2,6,4,4,5,4,6,6  ; 40
db 2,8,4,5,4,5,5,6,5,5,4,5,2,2,4,3  ; 50
db 2,8,4,5,3,4,3,6,2,6,4,4,5,4,5,5  ; 60
db 2,8,4,5,4,5,5,6,5,5,5,5,2,2,3,6  ; 70
db 2,8,4,5,3,4,3,6,2,6,5,4,5,2,4,5  ; 80
db 2,8,4,5,4,5,5,6,5,5,5,5,2,2,12,5 ; 90
db 3,8,4,5,3,4,3,6,2,6,4,4,5,2,4,4  ; A0
db 2,8,4,5,4,5,5,6,5,5,5,5,2,2,3,4  ; B0
db 3,8,4,5,4,5,4,7,2,5,6,4,5,2,4,9  ; C0
db 2,8,4,5,5,6,6,7,4,5,4,5,2,2,6,3  ; D0
db 2,8,4,5,3,4,3,6,2,4,5,3,4,3,4,3  ; E0
db 2,8,4,5,4,5,5,6,3,4,5,4,2,2,4,3  ; F0

%include "SPCflags.asm"

; This code should be copied into the top of the address space
SPC_ROM_CODE:
 db 0xCD,0xEF,0xBD,0xE8,0x00,0xC6,0x1D,0xD0
 db 0xFC,0x8F,0xAA,0xF4,0x8F,0xBB,0xF5,0x78
 db 0xCC,0xF4,0xD0,0xFB,0x2F,0x19,0xEB,0xF4
 db 0xD0,0xFC,0x7E,0xF4,0xD0,0x0B,0xE4,0xF5
 db 0xCB,0xF4,0xD7,0x00,0xFC,0xD0,0xF3,0xAB
 db 0x01,0x10,0xEF,0x7E,0xF4,0x10,0xEB,0xBA
 db 0xF6,0xDA,0x00,0xBA,0xF4,0xC4,0xF4,0xDD
 db 0x5D,0xD0,0xDB,0x1F,0x00,0x00,0xC0,0xFF

Read_Func_Map:              ; Mappings for SPC Registers
 dd SPC_READ_INVALID
 dd SPC_READ_INVALID
 dd SPC_READ_DSP_ADDR
 dd SPC_READ_DSP_DATA
 dd SPC_READ_PORT0R
 dd SPC_READ_PORT1R
 dd SPC_READ_PORT2R
 dd SPC_READ_PORT3R
 dd SPC_READ_INVALID
 dd SPC_READ_INVALID
 dd SPC_READ_INVALID
 dd SPC_READ_INVALID
 dd SPC_READ_INVALID
 dd SPC_READ_COUNTER_0
 dd SPC_READ_COUNTER_1
 dd SPC_READ_COUNTER_2

Write_Func_Map:             ; Mappings for SPC Registers
 dd SPC_WRITE_INVALID
 dd SPC_WRITE_CTRL
 dd SPC_WRITE_DSP_ADDR
 dd SPC_WRITE_DSP_DATA
 dd SPC_WRITE_PORT0W
 dd SPC_WRITE_PORT1W
 dd SPC_WRITE_PORT2W
 dd SPC_WRITE_PORT3W
 dd SPC_WRITE_INVALID
 dd SPC_WRITE_INVALID
 dd SPC_WRITE_TIMER_0
 dd SPC_WRITE_TIMER_1
 dd SPC_WRITE_TIMER_2
 dd SPC_WRITE_INVALID
 dd SPC_WRITE_INVALID
 dd SPC_WRITE_INVALID

ALIGND
offset_to_bit:  db 0x01,0x02,0x04,0x08,0x10,0x20,0x40,0x80
offset_to_not:  db 0xFE,0xFD,0xFB,0xF7,0xEF,0xDF,0xBF,0x7F

section .bss
ALIGND
SPC_Register_Base:

SPC_Code_Base:  skipl
EXPORT _PC ,skipl
EXPORT _YA
EXPORT _A  ,skipb
EXPORT _Y  ,skipb
            skipw
SPC_PAGE:   skipl
EXPORT _SP ,skipl

EXPORT _PSW,skipb   ; Processor status word
EXPORT _X  ,skipb

EXPORT_C SPC_PORT0R,skipb
EXPORT_C SPC_PORT1R,skipb
EXPORT_C SPC_PORT2R,skipb
EXPORT_C SPC_PORT3R,skipb

EXPORT_C SPC_PORT0W,skipb
EXPORT_C SPC_PORT1W,skipb
EXPORT_C SPC_PORT2W,skipb
EXPORT_C SPC_PORT3W,skipb

%ifdef DEBUG
SPC_TEMP_ADD:   skipl
%endif

section .text
ALIGNC
SNES_R2140_SPC: ; APUI00
 Execute_SPC SaveCycles
 mov al,[_SPC_PORT0W]
 ret

ALIGNC
SNES_R2141_SPC: ; APUI01
 Execute_SPC SaveCycles
 mov al,[_SPC_PORT1W]
 ret

ALIGNC
SNES_R2142_SPC: ; APUI02
 Execute_SPC SaveCycles
 mov al,[_SPC_PORT2W]
 ret

ALIGNC
SNES_R2143_SPC: ; APUI03
 Execute_SPC SaveCycles
 mov al,[_SPC_PORT3W]
 ret

ALIGNC
SNES_W2140_SPC: ; APUI00
 Execute_SPC SaveCycles
 mov [_SPC_PORT0R],al
 ret

ALIGNC
SNES_W2141_SPC: ; APUI01
 Execute_SPC SaveCycles
 mov [_SPC_PORT1R],al
 ret

ALIGNC
SNES_W2142_SPC: ; APUI02
 Execute_SPC SaveCycles
 mov [_SPC_PORT2R],al
 ret

ALIGNC
SNES_W2143_SPC: ; APUI03
 Execute_SPC SaveCycles
 mov [_SPC_PORT3R],al
 ret

ALIGNC
SPC_SET1:
 shr eax,5
 mov ebx,B_SPC_PAGE
 mov bl,[1+R_NativePC]
 add R_NativePC,byte 2
 mov ah,[offset_to_bit+eax]
 GET_BYTE_SPC     
 or al,ah
 SET_BYTE_SPC
 jmp SPC_BACK

ALIGNC
SPC_CLR1:
 shr eax,5
 mov ebx,B_SPC_PAGE
 mov bl,[1+R_NativePC]
 add R_NativePC,byte 2
 mov ah,[offset_to_not+eax]
 GET_BYTE_SPC     
 and al,ah
 SET_BYTE_SPC
 jmp SPC_BACK

ALIGNC
SPC_BBS:
 shr eax,5
 mov ebx,B_SPC_PAGE
 mov bl,[1+R_NativePC]
 mov ah,[offset_to_bit+eax]
 GET_BYTE_SPC
 test al,ah
 jz .not_taken
 movsx eax,byte [2+R_NativePC]
 add R_NativePC,eax
 add R_Cycles,byte 2    ; branch taken
.not_taken:
 add R_NativePC,byte 3
 jmp SPC_BACK

ALIGNC
SPC_BBC:
 shr eax,5
 mov ebx,B_SPC_PAGE
 mov bl,[1+R_NativePC]
 mov ah,[offset_to_bit+eax]
 GET_BYTE_SPC         
 test al,ah
 jnz .not_taken
 movsx eax,byte [2+R_NativePC]
 add R_NativePC,eax
 add R_Cycles,byte 2    ; branch taken
.not_taken:
 add R_NativePC,byte 3
 jmp SPC_BACK

ALIGNC
EXPORT_C Make_SPC
 pusha
 mov eax,SNES_R2140_SPC
 mov [Read_Map_21+0x40*4],eax
 mov eax,SNES_W2140_SPC
 mov [Write_Map_21+0x40*4],eax
 mov eax,SNES_R2141_SPC
 mov [Read_Map_21+0x41*4],eax
 mov eax,SNES_W2141_SPC
 mov [Write_Map_21+0x41*4],eax
 mov eax,SNES_R2142_SPC
 mov [Read_Map_21+0x42*4],eax
 mov eax,SNES_W2142_SPC
 mov [Write_Map_21+0x42*4],eax
 mov eax,SNES_R2143_SPC
 mov [Read_Map_21+0x43*4],eax
 mov eax,SNES_W2143_SPC
 mov [Write_Map_21+0x43*4],eax
 popa
 ret

ALIGNC
SPC_INVALID:
 mov [_Map_Byte],al     ; al contains opcode!

 xor eax,eax
 dec word [_PC]         ; Adjust address to correct for pre-increment
 mov eax,[_PC]
 mov [_Map_Address],eax ; this just sets the error output up correctly!

 mov al,R_PSW_B
 mov [_PSW],al

 jmp _InvalidSPCOpcode  ; This exits.. avoids conflict with other things!

ALIGNC
EXPORT_C Reset_SPC
 pusha

 cld

 ; This sets up the ROM code, done before registers as it
 ; contains the reset address for the ROM code (0xFFC0)
 mov esi,SPC_ROM_CODE
 mov edi,_SPCRAM+0xFFC0
 mov ecx,64/4
 push es
 push ds
 pop es
 rep movsd
 pop es

 ; Get ROM reset vector and setup Program Counter
 movzx eax,word [SPC_ROM_CODE+(0xFFFE-0xFFC0)]
 mov [_PC],eax

 mov eax,[SNES_Cycles]
 mov [SPC_last_cycles],eax

 ; Reset the sound DSP registers
 xor eax,eax
 mov edi,_SPC_DSP
 mov ecx,128/4
 push es
 push ds
 pop es
 rep stosd
 pop es

 mov [_SPC_Cycles],eax  ; Clear Cycle Count
 mov [_TotalCycles],eax
 mov [SPC_PAGE],eax     ; Used to save looking up P flag for Direct page stuff!
 mov dword [_SP],0x01FF ; Reset registers
 mov [_YA],eax
 mov [_X],al
 mov [_PSW],al          ; Clear Flags Register

 mov byte [_SPC_CTRL],0x80
 mov dword [SPC_FFC0_Address],SPC_ROM_CODE-0xFFC0
 mov dword [SPC_Code_Base],SPC_ROM_CODE-0xFFC0

 ; Reset timers
 mov [_SPC_CNT0],al
 mov [_SPC_CNT1],al
 mov [_SPC_CNT2],al
 mov dword [_SPC_TIMT0],256
 mov dword [_SPC_TIMT1],256
 mov dword [_SPC_TIMT2],256
 mov [_SPC_TIMER0],eax
 mov [_SPC_TIMER1],eax
 mov [_SPC_TIMER2],eax
 mov [_CycleLatch0],eax
 mov [_CycleLatch1],eax
 mov [_CycleLatch2],eax

 ; Reset SPC700 output ports
 mov [_SPC_PORT0W],al
 mov [_SPC_PORT1W],al
 mov [_SPC_PORT2W],al
 mov [_SPC_PORT3W],al

 dec eax
 ; Reset SPC700 input ports
 mov [_SPC_PORT0R],al
 mov [_SPC_PORT1R],al
 mov [_SPC_PORT2R],al
 mov [_SPC_PORT3R],al

 ; Reset sound DSP port address
 mov [_SPC_DSP_ADDR],al

 popa
 ret

SPC_SHOW_REGISTERS:
 pusha
;mov al,R_PSW_B
;mov [_PSW],al
 call _DisplaySPC
 popa
 ret

ALIGNC
EXPORT_C SPC_START
 mov al,[CPU_Executing]
 push eax
 mov byte [CPU_Executing],0

 LOAD_CYCLES
 LOAD_PSW
 LOAD_PC
 LOAD_BASE
 xor eax,eax
 jmp SPC_LOOP

ALIGNC
SPC_BACK:
;cmp R_Base,SPC_Register_Base
;jne 0b

%ifdef DEBUG
;mov ebx,[SPC_TEMP_ADD]
;mov [_OLD_SPC_ADDRESS],ebx
%endif
 cmp R_Cycles,0x80000000
global _SPC_Cycles
_SPC_Cycles equ $-4     ; Number of cycles to execute for SPC
%ifdef TRACKERS
 ja near SPC_OUT        ; Do another instruction if cycles left
%else
 ja SPC_OUT             ; Do another instruction if cycles left
%endif

SPC_LOOP:

; This code is for a SPC-tracker dump... #define TRACKERS to make a dump
; of the CPU state before each instruction - uncomment the calls to
; _Wangle__Fv and _exit to force the emulator to exit when the buffer
; fills. TRACKERS must be defined to the size of the buffer to be used -
; which must be a power of two, and the variables required by this and the
; write in Wangle() (romload.cc) exist only if DEBUG and SPCTRACKERS are
; also defined in romload.cc.
%ifdef TRACKERS
extern _SPC_LastIns
extern _SPC_InsAddress
extern _Wangle__Fv
extern _exit
 mov edi,[_SPC_LastIns]     ;
 add edi,[_SPC_InsAddress]  ;
 SAVE_PC eax                ;
 mov [edi],ah               ;
 mov [1+edi],al             ;
 mov al,[_A]                ;
 mov [2+edi],al             ;
 mov al,[_X]                ;
 mov [3+edi],al             ;
 mov al,[_Y]                ;
 mov [4+edi],al             ;
 mov al,[_SP]               ;
 mov [5+edi],al             ;
 SETUPFLAGS_SPC             ;
 mov [6+edi],al             ;

 mov al,[esi]               ;
 mov [7+edi],al             ;
 mov eax,[1+esi]            ;
 mov [8+edi],eax            ;
 mov eax,[5+esi]            ;
 mov [12+edi],eax           ;

 mov edi,[_SPC_LastIns]     ;
 add edi,byte 16            ;
 and edi,(TRACKERS-1)       ;
 mov [_SPC_LastIns],edi     ;
 test edi,edi               ;
 jnz .buffer_not_full       ;
 call _Wangle__Fv           ;
 jmp _exit                  ;
                            ;
.buffer_not_full:           ;
 xor eax,eax                ;
%endif

;mov ebx,[_PC]          ; PC now setup
;mov R_NativePC,[SPC_Code_Base]
;add R_NativePC,ebx
 xor eax,eax
%ifdef DEBUG
;mov [SPC_TEMP_ADD],ebx
%endif
 mov al,[R_NativePC]    ; Fetch opcode
 mov ebx,eax
;inc ebx
 mov bl,[SPCCycleTable+eax]
 add R_Cycles,ebx               ; Update cycle counter
 jmp dword [SPCOpTable+eax*4]   ; jmp to opcode handler

ALIGNC
SPC_OUT:
 SAVE_PC ebx
 SAVE_PSW
 SAVE_CYCLES   ; Set cycle counter

 pop eax
 mov [CPU_Executing],al
 ret                    ; Return to CPU emulation

%include "SPCaddr.ni"   ; Include addressing mode macros
%include "SPCmacro.ni"  ; Include instruction macros

%include "SPC0.ni"      ; Include opcodes 0x00-0x0F
%include "SPC1.ni"      ; Include opcodes 0x10-0x1F
%include "SPC2.ni"      ; Include opcodes 0x20-0x2F
%include "SPC3.ni"      ; Include opcodes 0x30-0x3F
%include "SPC4.ni"      ; Include opcodes 0x40-0x4F
%include "SPC5.ni"      ; Include opcodes 0x50-0x5F
%include "SPC6.ni"      ; Include opcodes 0x60-0x6F
%include "SPC7.ni"      ; Include opcodes 0x70-0x7F
%include "SPC8.ni"      ; Include opcodes 0x80-0x8F
%include "SPC9.ni"      ; Include opcodes 0x90-0x9F
%include "SPCA.ni"      ; Include opcodes 0xA0-0xAF
%include "SPCB.ni"      ; Include opcodes 0xB0-0xBF
%include "SPCC.ni"      ; Include opcodes 0xC0-0xCF
%include "SPCD.ni"      ; Include opcodes 0xD0-0xDF
%include "SPCE.ni"      ; Include opcodes 0xE0-0xEF
%include "SPCF.ni"      ; Include opcodes 0xF0-0xFF

SPC_READ_DSP_ADDR:
 mov al,[_SPC_DSP_ADDR]
 ret

SPC_READ_DSP_DATA:
 pusha
 call _SPC_READ_DSP
 popa
 push ebx
 mov bl,[_SPC_DSP_ADDR]
 and ebx,byte 0x7F
 mov al,[_SPC_DSP+ebx]    ; read from DSP register
 pop ebx
 ret

SPC_READ_PORT0R:
 mov al,[_SPC_PORT0R]
 ret
SPC_READ_PORT1R:
 mov al,[_SPC_PORT1R]
 ret
SPC_READ_PORT2R:
 mov al,[_SPC_PORT2R]
 ret
SPC_READ_PORT3R:
 mov al,[_SPC_PORT3R]
 ret

; WOOPS... TIMER registers are write only, the actual timer clock is internal not accessible!

; COUNTERS ARE 4 BIT, upon read they reset to 0 status

SPC_READ_COUNTER_0:
 pusha
 SAVE_CYCLES    ; Set cycle counter
 call _Update_SPC_Timer_0
 popa
 mov al,[_SPC_CNT0]
 mov byte [_SPC_CNT0],0
 ret

SPC_READ_COUNTER_1:
 pusha
 SAVE_CYCLES    ; Set cycle counter
 call _Update_SPC_Timer_1
 popa
 mov al,[_SPC_CNT1]
 mov byte [_SPC_CNT1],0
 ret

SPC_READ_COUNTER_2:
 pusha
 SAVE_CYCLES    ; Set cycle counter
 call _Update_SPC_Timer_2
 popa
 mov al,[_SPC_CNT2]
 mov byte [_SPC_CNT2],0
 ret

; | ---- | ---- | PC32 | PC10 | ---- | ST2  | ST1  | ST0  |
SPC_WRITE_CTRL:
 push eax
 mov ah,0
 test al,al     ; New for 0.25 - read hidden RAM
 mov edi,_SPCRAM
 jns .rom_disabled
 mov edi,SPC_ROM_CODE-0xFFC0

.rom_disabled:
 mov [SPC_FFC0_Address],edi

 test al,0x10       ; Reset ports 0/1 to 00 if set
 jz .no_clear_01
 mov [_SPC_PORT0R],ah   ; Ports read by SPC should be reset! 
 mov [_SPC_PORT1R],ah   ; Thanks to Butcha for fix!

.no_clear_01:
 test al,0x20       ; Reset ports 2/3 to 00 if set
 jz .no_clear_23
 mov [_SPC_PORT2R],ah
 mov [_SPC_PORT3R],ah

.no_clear_23:
 mov edi,R_Cycles
 test byte [_SPC_CTRL],4
 jnz .no_enable_timer_2
 test al,4
 jz  .no_enable_timer_2
 mov byte [_SPC_CNT2],0
 mov dword [_SPC_TIMER2],0
 mov [_CycleLatch2],edi

.no_enable_timer_2:
 test byte [_SPC_CTRL],2
 jnz .no_enable_timer_1
 test al,2
 jz .no_enable_timer_1
 mov byte [_SPC_CNT1],0
 mov dword [_SPC_TIMER1],0
 mov [_CycleLatch1],edi

.no_enable_timer_1:
 test byte [_SPC_CTRL],1
 jnz .no_enable_timer_0
 test al,1
 jz .no_enable_timer_0
 mov byte [_SPC_CNT0],0
 mov dword [_SPC_TIMER0],0
 mov [_CycleLatch0],edi

.no_enable_timer_0:
 pop eax
 mov [_SPC_CTRL],al
 ret

SPC_WRITE_DSP_ADDR:
 mov [_SPC_DSP_ADDR],al
 ret

SPC_WRITE_DSP_DATA:
 mov [_SPC_DSP_DATA],al
 pusha
 call _SPC_WRITE_DSP
 popa
 ret

SPC_WRITE_PORT0W:
 mov [_SPC_PORT0W],al
 ret
SPC_WRITE_PORT1W:
 mov [_SPC_PORT1W],al
 ret
SPC_WRITE_PORT2W:
 mov [_SPC_PORT2W],al
 ret
SPC_WRITE_PORT3W:
 mov [_SPC_PORT3W],al
 ret

SPC_WRITE_TIMER_0:
 pusha
 SAVE_CYCLES    ; Set cycle counter
 call _Update_SPC_Timer_0   ; Timer must catch up before changing target
 popa
 test al,al
 mov [_SPC_TIMT0],al    ; (0.32) Butcha - timer targets are writable
 setz [_SPC_TIMT0+1]    ; 0 = 256
 ret

SPC_WRITE_TIMER_1:
 pusha
 SAVE_CYCLES    ; Set cycle counter
 call _Update_SPC_Timer_1   ; Timer must catch up before changing target
 popa
 test al,al
 mov [_SPC_TIMT1],al    ; (0.32) Butcha - timer targets are writable
 setz [_SPC_TIMT1+1]    ; 0 = 256
 ret

SPC_WRITE_TIMER_2:
 pusha
 SAVE_CYCLES    ; Set cycle counter
 call _Update_SPC_Timer_2   ; Timer must catch up before changing target
 popa
 test al,al
 mov [_SPC_TIMT2],al    ; (0.32) Butcha - timer targets are writable
 setz [_SPC_TIMT2+1]    ; 0 = 256
 ret

section .text; ALIGNC
section .data; ALIGND
section .bss; ALIGND
